/**
 *############################################################################
 * A component of the Greenstone Librarian Interface, part of the Greenstone
 * digital library suite from the New Zealand Digital Library Project at the
 * University of Waikato, New Zealand.
 *
 * Author: Michael Dewsnip, NZDL Project, University of Waikato, NZ
 * Based on code by John Thompson
 *
 * Copyright (C) 2005 New Zealand Digital Library Project
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *############################################################################
 */

package org.greenstone.gatherer.gui;


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.tree.*;
import org.greenstone.gatherer.Configuration;
import org.greenstone.gatherer.DebugStream;
import org.greenstone.gatherer.Dictionary;
import org.greenstone.gatherer.Gatherer;
import org.greenstone.gatherer.collection.CollectionTree;
import org.greenstone.gatherer.collection.CollectionTreeNode;
import org.greenstone.gatherer.gui.tree.DragTree;
import org.greenstone.gatherer.metadata.MetadataElement;
import org.greenstone.gatherer.metadata.MetadataValue;
import org.greenstone.gatherer.metadata.MetadataValueTableEntry;
import org.greenstone.gatherer.metadata.MetadataValueTreeNode;
import org.greenstone.gatherer.util.DragGroup;
import org.greenstone.gatherer.util.TreeSynchronizer;


/** Provides a view of controls for the editing of metadata.
 */
public class EnrichPane
    extends JPanel
    implements TreeSelectionListener
{
    static private Dimension MINIMUM_SIZE = new Dimension(100, 100);
    static private Dimension COLLECTION_TREE_SIZE = new Dimension(250, 500);

    /** The collection tree. */
    private CollectionTree collection_tree = null;
    /** The currently reported selection. */
    private CollectionTreeNode[] file_nodes = null;
    /** Used to dynamically filter the collection tree. Synchronized with the same filter in the Gather view. */
    private Filter collection_filter = null;
    /** The label at the top of the collection tree, which shows the collection name. */
    private JLabel collection_label;
    /** The splitpane dividing the collection tree and the metadata editing controls. */
    private JSplitPane external_split;
    /** The metadata value table shows the metadata values that are currently assigned to a file. */
    private MetadataValueTablePane metadata_value_table_pane = null;
    /** The metadata value tree shows the metadata values that are currently assigned to a metadata element. */
    private MetadataValueTreePane metadata_value_tree_pane = null;
    /** Provide synchronization between the collection trees in this view and the collection pane view. */
    private TreeSynchronizer collection_tree_sync = null;


    /** Constructor.
     * @param tree_sync The <strong>TreeSynchronizer</strong> to be used on the collection tree
     */
    public EnrichPane(TreeSynchronizer collection_tree_sync)
    {
	this.collection_tree_sync = collection_tree_sync;

	// Create the metadata value tree pane
	metadata_value_tree_pane = new MetadataValueTreePane();
	metadata_value_tree_pane.addMetadataValueTreeSelectionListener(new MetadataValueTreeSelectionListener());

	// Create metadata value table pane
	metadata_value_table_pane = new MetadataValueTablePane();
	metadata_value_table_pane.addMetadataValueTableListSelectionListener(new MetadataValueTableListSelectionListener());
	metadata_value_table_pane.addMetadataValueTableMouseListener(new MetadataValueTableMouseListener());
	metadata_value_table_pane.addMetadataValueTextFieldDocumentListener(new MetadataValueTextFieldDocumentListener());
	metadata_value_table_pane.addMetadataValueTextFieldKeyListener(new MetadataValueTextFieldKeyListener());
    }


    /** Some actions can only occur after this panel has been displayed on-screen, so this method is provided to do exactly that. Such actions include the proportioning of the split panes and the setting of table column widths.
     */
    public void afterDisplay()
    {
	external_split.setDividerLocation(0.3);
    }


    /** Used to create, connect and layout the components to be shown on this control panel.
     * @see org.greenstone.gatherer.Gatherer
     * @see org.greenstone.gatherer.file.FileOpenActionListener
     * @see org.greenstone.gatherer.gui.Filter
     */
    public void display()
    {
	// Creation
	JPanel collection_pane = new JPanel(new BorderLayout());
	collection_pane.setMinimumSize(MINIMUM_SIZE);
	collection_pane.setPreferredSize(COLLECTION_TREE_SIZE);

	external_split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);

	collection_label = new JLabel();
	Dictionary.registerText(collection_label, "Collection.No_Collection");
	collection_label.setOpaque(true);

	DragGroup group = new DragGroup();
	collection_tree = new CollectionTree("Enrich", Gatherer.c_man.getCollectionTreeModel(), false);
	group.add(collection_tree);
	collection_tree.getSelectionModel().setSelectionMode(TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
	collection_tree.putClientProperty("JTree.lineStyle", "Angled");
	collection_tree.addMouseListener(Gatherer.g_man.foa_listener);
	collection_tree.addMouseListener(new RightButtonListener());
	collection_tree.addTreeSelectionListener(this);
	collection_tree.addTreeExpansionListener(Gatherer.g_man.foa_listener);
	collection_tree.setBackgroundNonSelectionColor(Configuration.getColor("coloring.collection_tree_background", false));
	collection_tree.setTextNonSelectionColor(Configuration.getColor("coloring.collection_tree_foreground", false));
	collection_tree.setBackgroundSelectionColor(Configuration.getColor("coloring.collection_selection_background", false));
	collection_tree.setTextSelectionColor(Configuration.getColor("coloring.collection_selection_foreground", false));
	collection_tree.setRootVisible(false);

	KeyListenerImpl key_listener = new KeyListenerImpl();
	collection_tree.addKeyListener(key_listener);
	JScrollPane collection_scroll = new JScrollPane(collection_tree);

	collection_filter = Gatherer.g_man.getFilter(collection_tree);
	collection_filter.setBackground(Configuration.getColor("coloring.collection_heading_background", false));
	collection_filter.setEditable(Configuration.getMode() > Configuration.LIBRARIAN_MODE);
	Dictionary.registerTooltip(collection_filter.getComboBox(), "Collection.Filter_Tooltip");

	// Layout
	collection_pane.setBorder(BorderFactory.createCompoundBorder(BorderFactory.createEmptyBorder(3,3,3,3), BorderFactory.createLoweredBevelBorder()));
	collection_pane.setMinimumSize(MINIMUM_SIZE);
	collection_pane.setPreferredSize(new Dimension(Gatherer.g_man.getSize().width / 3, Gatherer.g_man.getSize().height));

	// Collection Pane
	collection_pane.add(collection_label, BorderLayout.NORTH);
	collection_pane.add(collection_scroll, BorderLayout.CENTER);
	collection_pane.add(collection_filter, BorderLayout.SOUTH);

	JSplitPane metadata_editing_pane = new JSplitPane(JSplitPane.VERTICAL_SPLIT);
	metadata_editing_pane.setBorder(BorderFactory.createEmptyBorder(0,5,5,5));
	metadata_editing_pane.setDividerSize(8);

	metadata_editing_pane.add(metadata_value_table_pane, JSplitPane.TOP);
	metadata_editing_pane.add(metadata_value_tree_pane, JSplitPane.BOTTOM);
	metadata_editing_pane.setDividerLocation(250);

	external_split.add(collection_pane, JSplitPane.LEFT);
	external_split.add(metadata_editing_pane, JSplitPane.RIGHT);

	this.setLayout(new BorderLayout());
	this.add(external_split, BorderLayout.CENTER);
    }


    /** Called to inform this pane that it has just gained focus as an effect of the user clicking on its tab
     */
    public void gainFocus()
    {
	// If there is nothing in the collection tree there is nothing to do
	if (collection_tree == null || collection_tree.getRowCount() == 0) {
	    return;
	}

	// Select the first node in the tree if nothing is already selected
	if (collection_tree.getSelectionPaths() == null) {
 	    collection_tree.setImmediate(true);
 	    collection_tree.setSelectionRow(0);
 	    collection_tree.setImmediate(false);
	    return;
	}

	// Force all of the controls to be updated
	valueChanged(null);
    }


    /** Called whenever the detail mode changes to ensure the filters are at an appropriate level (ie only editable by those that understand regular expression matching)
     * @param mode the mode level as an int
     */
    public void modeChanged(int mode)
    {
	collection_filter.setEditable(mode > Configuration.LIBRARIAN_MODE);
    }


    /** Refresh this pane, depending on what has just happened (refresh_reason). */
    public void refresh(int refresh_reason, boolean collection_loaded)
    {
	if (collection_loaded) {
	    // Update collection label
	    Dictionary.registerText(collection_label, "Collection.Collection");
	    collection_label.setBackground(Configuration.getColor("coloring.collection_heading_background", false));
	    collection_label.setForeground(Configuration.getColor("coloring.collection_heading_foreground", false));

	    // Update collection tree
	    if (refresh_reason == Gatherer.COLLECTION_OPENED) {
		collection_tree.setModel(Gatherer.c_man.getCollectionTreeModel());
	    }

	    // Update collection filter
	    collection_filter.setBackground(Configuration.getColor("coloring.collection_heading_background", false));
	}
	else {
	    // Update collection label
	    Dictionary.registerText(collection_label, "Collection.No_Collection");
	    collection_label.setBackground(Color.lightGray);
	    collection_label.setForeground(Color.black);

	    // Update collection tree
	    collection_tree.setModel(new DefaultTreeModel(new DefaultMutableTreeNode("Error")));

	    // Update collection filter
	    collection_filter.setBackground(Color.lightGray);
	}

	// Enable or disable the controls
	collection_tree.setEnabled(collection_loaded);
	collection_filter.setEnabled(collection_loaded);

	// Ensure that this collection tree view is synchronized with all others
	collection_tree_sync.add(collection_tree);

	// Force the metadata table to be rebuilt (for switching extracted metadata on or off)
	if (refresh_reason == Gatherer.PREFERENCES_CHANGED) {
	    metadata_value_table_pane.stopEditingAndRebuild(file_nodes);
	}
    }


    /** Called whenever the collection tree selection changes. This causes the metadata value table to be rebuilt. */
    public void valueChanged(TreeSelectionEvent event)
    {
	// Nothing selected in the collection tree
	if (collection_tree.getSelectionCount() == 0) {
	    file_nodes = null;
	}

	// Some files selected in the collection tree
	else {
	    TreePath paths[] = collection_tree.getSelectionPaths();
	    file_nodes = new CollectionTreeNode[paths.length];
	    for (int i = 0; i < paths.length; i++) {
		file_nodes[i] = (CollectionTreeNode) paths[i].getLastPathComponent();
	    }
	}

	// Update the metadata value table (and consequently, the metadata value tree)
	metadata_value_table_pane.stopEditingAndRebuild(file_nodes);
    }


    private class MetadataValueTableListSelectionListener
	implements ListSelectionListener
    {
	public void valueChanged(ListSelectionEvent list_selection_event)
	{
	    // We only want to handle one event per selection, so wait for the value to stabilise
	    if (list_selection_event.getValueIsAdjusting()) {
		return;
	    }

	    // Update the metadata value tree for the current table selection
	    metadata_value_tree_pane.rebuild(metadata_value_table_pane.getSelectedMetadataValueTableEntry());
	}
    }


    private class MetadataValueTableMouseListener
	extends MouseAdapter
    {
	public void mouseClicked(MouseEvent mouse_event)
	{
	    // We're only interested in clicks on the inherited column
	    if (metadata_value_table_pane.isMouseEventForInheritedMetadataValueTableColumn(mouse_event) == false) {
		return;
	    }

	    // If the selected metadata is inherited, switch to the folder it came from
	    MetadataValueTableEntry selected_metadata_value_table_entry = metadata_value_table_pane.getSelectedMetadataValueTableEntry();
	    if (selected_metadata_value_table_entry.isInheritedMetadata()) {
		collection_tree.setSelection(selected_metadata_value_table_entry.getFolderMetadataInheritedFrom());
	    }
	}
    }


    private class MetadataValueTextFieldDocumentListener
	implements DocumentListener
    {
	/** Gives notification that an attribute or set of attributes changed */
	public void changedUpdate(DocumentEvent document_event) {
	    validate(document_event);
	}

	/** Gives notification that there was an insert into the document */
	public void insertUpdate(DocumentEvent document_event) {
	    validate(document_event);
	}

	/** Gives notification that a portion of the document has been removed */
	public void removeUpdate(DocumentEvent document_event) {
	    validate(document_event);
	}


	/** Ensures that the value tree is updated in response to changes in the value text field */
	private void validate(DocumentEvent document_event)
	{
	    try {
		Document document = document_event.getDocument();
		String metadata_value_string = document.getText(0, document.getLength());
		metadata_value_tree_pane.selectBestPathForMetadataValue(metadata_value_string);
	    }
	    catch (Exception exception) {
		DebugStream.printStackTrace(exception);
	    }
	}
    }


    private class MetadataValueTextFieldKeyListener
	extends KeyAdapter
    {
	/** Gives notification of key events on the text field */
	public void keyPressed(KeyEvent key_event)
	{
	    // Tab: Auto-complete what is selected in the metadata value tree
	    if (key_event.getKeyCode() == KeyEvent.VK_TAB) {
		MetadataValueTreeNode selected_metadata_value_tree_node = metadata_value_tree_pane.getSelectedMetadataValueTreeNode();
		if (selected_metadata_value_tree_node != null) {
		    metadata_value_table_pane.setMetadataValueTextFieldValue(selected_metadata_value_tree_node.getFullValue());
		}

		// We do not want this event to be processed by the table also
		key_event.consume();
	    }

	    // Enter: save the current value then add a blank row for the selected metadata element
	    if (key_event.getKeyCode() == KeyEvent.VK_ENTER) {
		metadata_value_table_pane.stopEditingAndAddBlankRowForSelectedMetadataElement();
	    }
	}
    }


    private class MetadataValueTreeSelectionListener
	implements TreeSelectionListener
    {
	public void valueChanged(TreeSelectionEvent tree_selection_event)
	{
	    // When a node is selected in the tree, fill the metadata value text field with the selected node's value
	    MetadataValueTreeNode selected_metadata_value_tree_node = metadata_value_tree_pane.getSelectedMetadataValueTreeNode();
	    if (selected_metadata_value_tree_node != null) {
		metadata_value_table_pane.setMetadataValueTextFieldValue(selected_metadata_value_tree_node.getFullValue());
	    }
	}
    }


    /** This class listens for certain key presses, such as [Up] or [Down], -copied from Gather Pane */
    private class KeyListenerImpl
	extends KeyAdapter {
	private boolean vk_left_pressed = false;
	/** Called whenever a key that was pressed is released, it is this action that will cause the desired effects (this allows for the key event itself to be processed prior to this listener dealing with it). */
	public void keyReleased(KeyEvent event) {
	    ///ystem.err.println("Key Release detected. " + event.getKeyCode());
	    if (event.getKeyCode() == KeyEvent.VK_UP || event.getKeyCode() == KeyEvent.VK_DOWN) {
		DragTree tree = (DragTree)event.getSource();
		// we need to manually do the up and down selections
		boolean up = (event.getKeyCode() == KeyEvent.VK_UP);
		int current_row = tree.getLeadSelectionRow();
		tree.setImmediate(true);
		if (up) {
		    if (current_row > 0) {
			tree.clearSelection();
			tree.setSelectionRow(current_row-1);
		    }
		} else {
		    if (current_row < tree.getRowCount()-1){
			tree.clearSelection();
			tree.setSelectionRow(current_row+1);
		    }
		}
		tree.setImmediate(false);
	    }
	    else if (event.getKeyCode() == KeyEvent.VK_LEFT) {
		// left key on a file shifts the selection to the parent folder
		DragTree tree = (DragTree)event.getSource();
		TreePath path = tree.getLeadSelectionPath();
		if(path != null) {
		    File file = ((CollectionTreeNode) path.getLastPathComponent()).getFile();
		    if(file != null) {
			if (file.isFile() || vk_left_pressed) {
			    vk_left_pressed = false;
			    TreePath parent_path = path.getParentPath();
			    if (parent_path != null && parent_path.getParentPath() != null) {
				// if this file is in the top level folder, don't move the focus 
				tree.setImmediate(true);
				tree.clearSelection();
				tree.setSelectionPath(parent_path);
				tree.setImmediate(false);
			    }
			}
		    }
		}
	    }	    
	}

	// we need to watch for left clicks on an unopened folder - should shift the focus to teh parent folder. But because there is some other mysterious key listener that does opening and closing folders on right and left clicks, we must detect the situation before the other handler has done its job, and process it after. 
	public void keyPressed(KeyEvent event) {
	    if (event.getKeyCode() == KeyEvent.VK_LEFT) {
		// left key on  closed directory shifts the selection to the parent folder
		DragTree tree = (DragTree)event.getSource();
		TreePath path = tree.getLeadSelectionPath();
		if(path == null) return; 
		File file = ((CollectionTreeNode) path.getLastPathComponent()).getFile();
		if(file == null) return;
		
		if (file.isDirectory() && tree.isCollapsed(path)) {
		    vk_left_pressed = true;
		}		
	    }
	}
    }


    /** A listener for right mouse button clicks over the collection tree. */
    private class RightButtonListener
	extends MouseAdapter {
	/** Called whenever a mouse click occurs (right or left) over a target component.
	 * @param event A <strong>MouseEvent</strong> containing further information about the mouse click action.
	 * @see org.greenstone.gatherer.gui.EnrichPane.RightButtonMenu
	 */
	public void mouseClicked(MouseEvent event) {
	    if (SwingUtilities.isRightMouseButton(event)) {
		new RightButtonMenu(event);
	    }
	}
    }


    /** When a user right-clicks within the collection tree they are presented with a small popup menu of context based options. This class provides such functionality.
     */
    private class RightButtonMenu
	extends JPopupMenu
	implements ActionListener {

	/** The tree over which the right click action occurred. */
	private DragTree tree = null;
	/** The tree nodes selected when the right click action occurred. */
	private TreePath[] selection_paths = null;
	/** The file record over which the right click action occurred. */
	private CollectionTreeNode node = null;

	private JMenuItem collapse_folder = null;
	private JMenuItem expand_folder = null;
	private JMenuItem metaaudit = null;
	private JMenuItem open_externally = null;


	private RightButtonMenu(MouseEvent event)
	{
	    super();
	    this.tree = collection_tree;

	    // Note we have to use setImmediate() with the set selction paths
	    // otherwise the selection doesn't get updated until after the 
	    // popup comes up.

	    // the right click position
	    TreePath right_click_path = tree.getPathForLocation(event.getX(), event.getY());
	    if (right_click_path == null) {
		// user has clicked outside of the tree, clear the selection
		selection_paths = null;
		tree.setImmediate(true);
		tree.clearSelection();
		tree.setImmediate(false);
	    } else {
		// Get the paths currently selected in the tree
		selection_paths = tree.getSelectionPaths();
		if (selection_paths == null) {
		    // nothing currently selected - we shift the selection to 
		    // the node that was right clicked on
		    selection_paths = new TreePath[1];
		    selection_paths[0] = right_click_path;
		    tree.setImmediate(true);
		    tree.setSelectionPath(right_click_path);
		    tree.setImmediate(false);
		} else if (selection_paths.length == 1 && ! selection_paths[0].equals( right_click_path)) {
		    tree.setImmediate(true);
		    tree.clearSelection();
		    tree.setSelectionPath(right_click_path);
		    tree.setImmediate(false);
		    selection_paths[0] = right_click_path;
		} else {
		    // we had multiply selected paths in the tree.
		    // if we clicked on one of those paths, then use all the 
		    // current selection, otherwise clear the selection and 
		    // select the one we right clicked on
		    boolean clicked_in_selection = false;
		    for (int i=0; i<selection_paths.length; i++) {
			if (selection_paths[i].equals(right_click_path)) {
			    clicked_in_selection = true;
			    break;
			}
		    }
		    if (!clicked_in_selection) {
			// want the tree to update right away
			tree.setImmediate(true);
			tree.clearSelection();
			tree.setSelectionPath(right_click_path);
			tree.setImmediate(false);
			selection_paths = new TreePath[1];
			selection_paths[0] = right_click_path;
		    }
		}
	    }
	    // finally we have the correct selection paths!

	    // Create an appropriate context menu, based on what is selected
	    buildContextMenu(selection_paths);

	    // Show the popup menu on screen
	    show(tree, event.getX(), event.getY());
	}


	private void buildContextMenu(TreePath[] selection_paths)
	{
	    // If nothing is selected, no options are available
	    if (selection_paths == null) {
		return;
	    }

	    DebugStream.println("Number of files/folders selected: " + selection_paths.length);

	    // Always have meta-audit option
	    metaaudit = new JMenuItem(Dictionary.get("Menu.Metadata_View", collection_tree.getSelectionDetails()), KeyEvent.VK_A);
	    metaaudit.addActionListener(this);
	    add(metaaudit);

	    // Only meta-audit is available if multiple items are selected...
	    if (selection_paths.length > 1) {
		return;
	    }

	    TreePath path = selection_paths[0];
	    node = (CollectionTreeNode) path.getLastPathComponent();

	    // ---- Options for file nodes ----
	    if (node.isLeaf()) {
		// Open the file in an external program
		open_externally = new JMenuItem(Dictionary.get("Menu.Open_Externally"), KeyEvent.VK_O);
		open_externally.addActionListener(this);
		add(open_externally);
		return;
	    }

	    // ---- Options for folder nodes ----
	    // Collapse or expand, depending on current status
	    if (tree.isExpanded(path)) {
		collapse_folder = new JMenuItem(Dictionary.get("Menu.Collapse"), KeyEvent.VK_C);
		collapse_folder.addActionListener(this);
		add(collapse_folder);
	    }
	    else {
		expand_folder = new JMenuItem(Dictionary.get("Menu.Expand"), KeyEvent.VK_O);
		expand_folder.addActionListener(this);
		add(expand_folder);
	    }
	}


	/** Called whenever one of the menu items is clicked, this method then causes the appropriate effect. */
	public void actionPerformed(ActionEvent event)
	{
	    Object source = event.getSource();

	    // Collapse folder
	    if (source == collapse_folder) {
		tree.collapsePath(selection_paths[0]);
	    }

	    // Expand folder
	    else if (source == expand_folder) {
		tree.expandPath(selection_paths[0]);
	    }

	    // Meta-audit
	    else if (source == metaaudit) {
		Gatherer.g_man.showMetaAuditBox();
	    }

	    // Open in external program
	    else if (source == open_externally) {
		Gatherer.self.spawnApplication(node.getFile());
	    }
	}
    }
}
